#!/usr/bin/env python3


from charmhelpers.fetch import apt_install

from charmhelpers.core import hookenv, host

from charmhelpers.core.templating import render

from charms.reactive import (
    main,
    is_state,
    remove_state,
    set_state,
)

from charms.reactive.decorators import (
    hook,
    when,
    when_not,
)

import os
import subprocess


#
# Module configuration - these shouldn't need adjusting
#

rsync_config = '/etc/rsync-juju.d/ntplog.conf'
firewall_script = '/etc/network/ntplog.sh'

# All previous or current names by which the script symlink might have been known
firewall_remove = (
    '/etc/network/if-pre-up.d/firewall.sh',
    '/etc/network/if-pre-up.d/firewall',
    '/etc/network/if-pre-up.d/ntplog',
    '/etc/network/if-up.d/ntplog',
)

# Symlinks to the log script to make it survive reboots
firewall_link = (
    '/etc/network/if-up.d/ntplog',
)


#
# Utility functions
#
def status(status, msg):
    hookenv.log('%s: %s' % (status, msg))
    hookenv.status_set(status, msg)


def active(msg):
    status('active', msg)


def blocked(msg):
    status('blocked', msg)


def maint(msg):
    status('maintenance', msg)


def is_firewall_configured():
    """
    Return true if our main firewall rule is present
    in the INPUT chain for both IPv4 and IPv6.
    """
    check_cmds = (
        "iptables --check INPUT -j ntplog -p udp --dport 123",
        "ip6tables --check INPUT -j ntplog -p udp --dport 123",
    )

    for cmd in check_cmds:
        try:
            subprocess.check_call(cmd.split())
        except subprocess.CalledProcessError as e:
            hookenv.log('%s failed with return code %d' % (cmd, e.returncode))
            return e.returncode == 0

    # check_cmds succeeded
    return True


#
# Hooks
#


@hook('update-status')
def update_hook():
    """
    The firewall script is lightweight and should have no effect on a running system.
    Run it periodically in case something else is adjusting iptables rules.
    """
    if is_state('ntplog.installed') and not is_firewall_configured():
        start_firewall()


# layer-basic does not fire config.changed unconditionally on charm
# upgrade, which means we need to make that happen ourselves, in
# case there is a configuration-affecting change.

@hook('upgrade-charm')
def upgrade_charm():
    maint('Forcing reconfiguration on upgrade-charm')
    remove_state('ntplog.configured')


#
# Helper functions
#


def write_firewall_config():
    maint('configuring firewall')

    proto = "udp"
    port = 123
    log = "NTP"

    # create firewall script
    hookenv.log('rendering template to %s' % (firewall_script,))
    context = {
        'log': log,
        'port': port,
        'proto': proto,
    }
    render(
        source='firewall.j2',
        target=firewall_script,
        context=context,
        perms=0o755,
    )

    # create startup link
    try:
        for s in firewall_remove:
            if os.path.islink(s):
                os.remove(s)
                hookenv.log('removed old symlink %s' % (s,))
    except OSError as e:
        print(e)
        blocked('problem removing firewall symlink %s' % (e.filename,))
    try:
        for s in firewall_link:
            os.symlink(firewall_script, s)
            hookenv.log('created firewall symlink %s' % (s,))
    except OSError as e:
        print(e)
        blocked('problem creating firewall symlink %s' % (e.filename,))

    # force a restart now that we've changed the config
    remove_state('ntplog.started')
    hookenv.log('created firewall config')


def write_rsync_config(hosts_allow):
    maint('configuring rsync')
    context = {
        'hosts_allow': hosts_allow,
    }
    render(
        source='rsyncd.conf.j2',
        target=rsync_config,
        context=context,
        perms=0o644,
    )
    hookenv.open_port(873)
    hookenv.log('created rsync configuration')


#
# Reactive state handlers
#

@when('config.changed')
def config_changed():
    remove_state('ntplog.configured')


@when_not('ntplog.installed')
def install():
    """
    Transition from uninstalled to installed state by installing package dependencies.
    """
    maint('installing dependencies')
    apt_install(['ipset', 'rsync'])
    set_state('ntplog.installed')
    active('dependencies installed')


@when('ntplog.installed')
@when_not('ntplog.configured')
def write_configuration():
    """
    Transition from installed to configured state by writing the configuration.
    """
    config = hookenv.config()
    write_rsync_config(config['hosts-allow'])
    write_firewall_config()

    # change rsyslog to rotate logs daily, keep for a week, and not delay compression
    adjust_logrotate = [
        'sed',
        '-i',
        '-re',
        's/rotate [0-9]*$/rotate 7/; s/(weekly|monthly)/daily/; /^[^#]*delaycompress/d',
        '/etc/logrotate.d/rsyslog',
    ]
    try:
        subprocess.check_call(adjust_logrotate)
    except subprocess.CalledProcessError as e:
        blocked('%s failed with return code %d' % (adjust_logrotate, e.returncode))
        return

    # ensure kernel logging occurs only to kern.log
    render(
        source='00-ntplog.conf',
        target='/etc/rsyslog.d/00-ntplog.conf',
        context={},
        perms=0o644,
    )
    try:
        host.service_restart('rsyslog')
    except:
        blocked('rsyslog restart failed - please check unit log')
        remove_state('ntplog.configured')
        return

    set_state('ntplog.configured')
    active('ntplog configured')


@when('ntplog.configured')
@when_not('ntplog.started')
def start_firewall():
    """
    Transition from configured to started state by starting the firewall.
    """
    maint('starting firewall')
    if os.path.isfile(firewall_script):
        try:
            subprocess.check_call(firewall_script)
            set_state('ntplog.started')
            active('started firewall')
        except subprocess.CalledProcessError as e:
            blocked('%s failed with return code %d' % (firewall_script, e.returncode))
    else:
        remove_state('ntplog.configured')
        blocked('%s does not exist' % (firewall_script,))


if __name__ == '__main__':
    main()
